# coding=utf-8
from __future__ import unicode_literals

from datetime import date, datetime
from decimal import Decimal

import pytest
import tzlocal

from piecash import GnucashException, Commodity
from piecash.core import factories
from test_helper import db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic, is_not_on_web

# dummy line to avoid removing unused symbols

a = db_sqlite_uri, db_sqlite, new_book, new_book_USD, book_uri, book_basic


class TestFactoriesCommodities(object):
    def test_create_stock_accounts_simple(self, book_basic):
        with pytest.raises(GnucashException):
            factories.create_stock_accounts(book_basic.default_currency,
                                            broker_account=book_basic.accounts(name="broker"))

        broker = book_basic.accounts(name="broker")
        appl = Commodity(namespace="NMS", mnemonic="AAPL", fullname="Apple")
        acc, inc_accounts = factories.create_stock_accounts(appl,
                                                            broker_account=broker)

        assert inc_accounts == []
        assert broker.children == [acc]

    def test_create_stock_accounts_incomeaccounts(self, book_basic):
        broker = book_basic.accounts(name="broker")
        income = book_basic.accounts(name="inc")

        appl = Commodity(namespace="NMS", mnemonic="AAPL", fullname="Apple")
        appl["quoted_currency"] = "USD"
        acc, inc_accounts = factories.create_stock_accounts(appl,
                                                            broker_account=broker,
                                                            income_account=income,
                                                            income_account_types="D")
        assert len(inc_accounts) == 1

        acc, inc_accounts = factories.create_stock_accounts(appl,
                                                            broker_account=broker,
                                                            income_account=income,
                                                            income_account_types="CL")
        assert len(inc_accounts) == 1
        acc, inc_accounts = factories.create_stock_accounts(appl,
                                                            broker_account=broker,
                                                            income_account=income,
                                                            income_account_types="CS")
        assert len(inc_accounts) == 1
        acc, inc_accounts = factories.create_stock_accounts(appl,
                                                            broker_account=broker,
                                                            income_account=income,
                                                            income_account_types="I")
        assert len(inc_accounts) == 1
        acc, inc_accounts = factories.create_stock_accounts(appl,
                                                            broker_account=broker,
                                                            income_account=income,
                                                            income_account_types="D/CL/CS/I")
        assert len(income.children) == 4
        assert sorted(income.children, key=lambda x: x.guid) == sorted([_acc.parent for _acc in inc_accounts], key=lambda x: x.guid)
        assert broker.children == [acc]

    def test_create_stock_from_symbol(self, book_basic):
        if is_not_on_web():
            return
        factories.create_stock_from_symbol("AAPL", book_basic)

    def test_create_currency_from_ISO(self, book_basic):
        assert factories.create_currency_from_ISO("CAD").fullname == "Canadian Dollar"

        with pytest.raises(ValueError):
            factories.create_currency_from_ISO("EFR").fullname


class TestFactoriesTransactions(object):
    def test_single_transaction(self, book_basic):
        today = datetime.today()
        factories.single_transaction(today,
                                     today,
                                     "my test",
                                     Decimal(100),
                                     from_account=book_basic.accounts(name="inc"),
                                     to_account=book_basic.accounts(name="asset"))
        book_basic.save()
        tr = book_basic.transactions(description="my test")
        assert len(tr.splits) == 2
        sp1, sp2 = tr.splits
        if sp1.value > 0:
            sp2, sp1 = sp1, sp2
        # sp1 has negative value
        assert sp1.account == book_basic.accounts(name="inc")
        assert sp2.account == book_basic.accounts(name="asset")
        assert sp1.value == -sp2.value
        assert sp1.quantity == sp1.value
        assert tr.enter_date == tzlocal.get_localzone().localize(today.replace(microsecond=0))
        assert tr.post_date == tzlocal.get_localzone().localize(today.replace(hour=11, minute=0, second=0, microsecond=0))

    def test_single_transaction_tz(self, book_basic):
        today = tzlocal.get_localzone().localize(datetime.today())
        factories.single_transaction(today,
                                     today,
                                     "my test",
                                     Decimal(100),
                                     from_account=book_basic.accounts(name="inc"),
                                     to_account=book_basic.accounts(name="asset"))
        book_basic.save()
        tr = book_basic.transactions(description="my test")
        assert tr.post_date == today.replace(hour=11, minute=0, second=0, microsecond=0)
        assert tr.enter_date == today.replace(microsecond=0)

    def test_single_transaction_rollback(self, book_basic):
        today = tzlocal.get_localzone().localize(datetime.today())
        factories.single_transaction(today,
                                     today,
                                     "my test",
                                     Decimal(100),
                                     from_account=book_basic.accounts(name="inc"),
                                     to_account=book_basic.accounts(name="asset"))
        book_basic.validate()
        assert len(book_basic.transactions) == 1
        book_basic.cancel()
        assert len(book_basic.transactions) == 0



from __future__ import unicode_literals
from .core import Transaction, Account, Commodity, Price, Book

def attach_ledger(cls):
    def _process(fct):
        cls.__ledger__ = fct
    return _process

@attach_ledger(Transaction)
def ledger(tr):
    """Return a ledger-cli alike representation of the transaction"""
    s = ["{:%Y/%m/%d} * {}\n".format(tr.post_date, tr.description)]
    if tr.notes:
        s.append("\t;{}\n".format(tr.notes))
    for split in tr.splits:
        s.append("\t{:40} ".format(split.account.fullname))
        if split.account.commodity != tr.currency:
            s.append("{:10.2f} {} @@ {:.2f} {}".format(
                split.quantity, split.account.commodity.mnemonic, abs(split.value),
                tr.currency.mnemonic))
        else:
            s.append("{:10.2f} {}".format(split.value, tr.currency.mnemonic))
        if split.memo:
            s.append(" ;   {:20}".format(split.memo))
        s.append("\n")
    return "".join(s)


def format_commodity(commodity):
    mnemonic = commodity.mnemonic
    try:
        if mnemonic.encode('ascii').isalpha():
            return mnemonic
    except:
        pass
    return "\"{}\"" .format(mnemonic)  # TODO: escape " char in mnemonic


@attach_ledger(Commodity)
def ledger(cdty):
    """Return a ledger-cli alike representation of the commodity"""
    if cdty.mnemonic == "":
        return
    res = "commodity {}\n" .format(format_commodity(cdty))
    if cdty.fullname != "":
        res += "\tnote {}\n" .format(cdty.fullname)
    res += "\n"
    return res

@attach_ledger(Account)
def ledger(acc):
    """Return a ledger-cli alike representation of the account"""
    # ignore "dummy" accounts
    if acc.type is None or acc.parent is None:
        return
    if str(acc.commodity) == "template":
        return
    res = "account {}\n" .format(acc.fullname, )
    if acc.description != "":
        res += "\tnote {}\n" .format(acc.description,)

    res += "\tcheck commodity == \"{}\"\n" .format(acc.commodity.mnemonic)
    return res

@attach_ledger(Price)
def ledger(price):
    """Return a ledger-cli alike representation of the price"""
    return "P {:%Y/%m/%d %H:%M:%S} {} {} {}\n" .format(price.date, format_commodity(price.commodity), price.value, format_commodity(price.currency))

@attach_ledger(Book)
def ledger(book):
    """Return a ledger-cli alike representation of the book"""
    res = []

    # Commodities
    for commodity in book.commodities:
        res.append(ledger(commodity))

    # Accounts
    for acc in book.accounts:
        res.append(ledger(acc))
        res.append("\n")

    # Prices
    for price in sorted(book.prices, key=lambda x:x.date):
        res.append(ledger(price))
    res.append("\n")

    for trans in sorted(book.transactions,key=lambda x: x.post_date):
        res.append(ledger(trans))
        res.append("\n")

    return "".join(res)

def ledger(obj):
    return obj.__ledger__()



#!/usr/local/bin/python
"""Basic script to export QIF. Heavily untested ..."""
import argparse
from piecash import open_book, Transaction

try:
    import qifparse.qif as qif
except ImportError:
    print "You need to install the qifparse module ('pip install qifparse')"
# https://github.com/jemmyw/Qif/blob/master/QIF_references

parser = argparse.ArgumentParser(description="Generate a QIF representation of a gnucash book")
parser.add_argument("gnucash_filename",
                    help="the name of the gnucash file to process")
args = parser.parse_args()

map_gnc2qif = {
    "CASH": 'Cash',
    "BANK": 'Bank',
    "RECEIVABLE": 'Bank',
    "PAYABLE": 'Ccard',
    "MUTUAL": 'Bank',
    "CREDIT": 'Ccard',
    "ASSET": 'Oth A',
    "LIABILITY": 'Oth L',
    "TRADING": 'Oth L',
    # 'Invoice',  # Quicken for business only
    "STOCK": 'Invst',
}

with open_book(args.gnucash_filename, open_if_lock=True) as s:
    b = qif.Qif()
    map = {}
    for acc in s.accounts:
        if acc.parent == s.book.root_template: continue
        elif acc.type in ["INCOME", "EXPENSE", "EQUITY"]:
            item = qif.Category(name=acc.fullname,
                                description=acc.description,
                                expense=acc.type == "EXPENSE",
                                income=acc.type == "INCOME" or acc.type == "EQUITY",
            )
            b.add_category(item)
        elif acc.type in map_gnc2qif:
            actype = map_gnc2qif[acc.type]
            if actype=="Invst":
                # take parent account
                item = qif.Account(name=acc.fullname, account_type=actype)
            else:
                item = qif.Account(name=acc.fullname, account_type=actype)
            b.add_account(item)
        else:
            print("unknow {} for {}".format(acc.type, acc.fullname))

        map[acc.fullname] = item

    # print(str(b))
    def split_type(sp):
        qif_obj = map[sp.account.fullname]
        if isinstance(qif_obj, qif.Account):
            return qif_obj.account_type
        else:
            return "Expense" if qif_obj.expense else "Income"

    def sort_split(sp):
        type = split_type(sp)
        if type=="Invst":
            return 1
        elif type in ["Expense","Income"]:
            return 2
        else:
            return 0

    tpl = s.book.root_template

    for tr in s.transactions:
        if not tr.splits or len(tr.splits)<2: continue # split empty transactions
        if tr.splits[0].account.parent==tpl: continue # skip template transactions

        assert isinstance(tr, Transaction)

        splits = sorted(tr.splits, key=sort_split)
        if all(sp.account.commodity.namespace == "CURRENCY" for sp in splits):

            sp1, sp2 = splits[:2]
            item = qif.Transaction(date=tr.post_date,
                                   num=tr.num,
                                   payee=tr.description,
                                   amount=sp1.value,
                                   memo=sp1.memo,
            )
            if isinstance(map[sp2.account.fullname], qif.Account):
                item.to_account = sp2.account.fullname
            else:
                item.category = sp2.account.fullname

            if len(splits) > 2:
                for sp in splits[1:]:
                    if isinstance(map[sp.account.fullname], qif.Account):
                        asp = qif.AmountSplit(amount=-sp.value,
                                              memo=sp.memo,
                                              to_account=sp.account.fullname,
                        )
                    else:
                        asp = qif.AmountSplit(amount=-sp.value,
                                              memo=sp.memo,
                                              category=sp.account.fullname,
                        )
                    item.splits.append(asp)
            map[sp1.account.fullname].add_transaction(item,
                                                      header="!Type:{}".format(map[sp1.account.fullname].account_type))
        else:
            # match pattern of splits for an investment

            sp_account, sp_security, sp_others = splits[0], splits[1], splits[2:]

            assert split_type(sp_account) in ["Bank", "Cash"]
            assert split_type(sp_security) in ["Invst"]
            assert all(split_type(sp)=="Expense" for sp in sp_others)
            assert sp_security.account.parent.type=="BANK", "Security account has no parent STOCK account (aka a Brokerage account)"
            item = qif.Investment(date=tr.post_date,
                                  action="Buy" if sp_security.quantity>0 else "Sell",
                                  security=sp_security.account.commodity.mnemonic,
                                  price=sp_security.value / sp_security.quantity,
                                  quantity=sp_security.quantity,
                                  amount=sp_security.value,
                                  commission=sum(sp.value for sp in sp_others),
                                  first_line=tr.description,
                                  to_account=sp_account.account.fullname,
                                  amount_transfer=abs(sp_account.value),
                                  ) 
            map[sp_security.account.fullname]\
                .add_transaction(item, header="!Type:{}".format(split_type(sp_security)))

print(str(b))
